/*
 *  linux/kernel/lids.c
 *
 *  Author : Huagang Xie, (xie@gnuchina.org) 
 *
 *	This file contain the lids security options.
 */

/*
 *  Changes:
 *	
 *	[Oct 14 1999, Xie Hua Gang] initial creation 
 *	[Oct 17 1999, Xie Hua Gang] Added to do_truncate,do_link(). 
 *	[Oct 27 1999, Xie Hua Gang] Added Append Only ,Device MBR protection. 
 *      [Dec 05 1999, Philippe Biondi] New design. Added hidding process support. Added /dev/[k]mem protection.
 *      [Dec 08 1999, Xie Hua Gang] Added firewall rules protection.
 *      [Dec 09 1999, Philippe Biondi] Added ptrace protection. Permission flags for /dev/mem.
 *      [Dec 17 1999, Philippe Biondi] Added raw disk access, ioperm, iopl protection.
 *      [Jan 08 2000, Philippe Biondi] Use of capabilities bounding set.
 *
 * 	[May 07 2000, Xie Huagang] Add ACL in kernel, reference monitor.
 * 	
 * 	[Sep 26 2000, Xie Huagang] Port to linux 2.4.0-test8
 *
 * 	[Jan 10 2001, Xie Huagang] LIDS for Linux 2.4.0 released!
 *
 * 	[Jun 29 2001, Xie Huagang] Add time restrition support.
 */

#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/quotaops.h>

#include <asm/uaccess.h>
#include <asm/unaligned.h>
#include <asm/semaphore.h>
#include <asm/page.h>
#include <asm/pgtable.h>

#include <asm/namei.h>

#include <linux/utime.h>
#include <linux/file.h>
#include <linux/fs.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/config.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/sysctl.h>
#include <linux/rmd160.h>

#include <linux/lids.h>

/***********************************************************************
 *********************************************************************** 
 *
 *  Here are some consts and type definitions
 *
 ***********************************************************************
 ***********************************************************************/

#define LIDS_CONF_FILE	"/etc/lids/lids.conf"	/* the configuration file */
#define LIDS_CONF_DIR "/etc/lids"
#define LIDS_ADM_PATH "/sbin/lidsadm"              /* the adminstration tool */
#define LIDS_PW_FILE "/etc/lids/lids.pw"
#define LIDS_PW_LEN  32
#define LIDS_CAP_FILE "/etc/lids/lids.cap"

typedef struct
{
int last_secure;
int last_s_acl;
int last_o_acl;
unsigned long search_secure;
unsigned long search_s_acl;
unsigned long fastguess[2048];
struct secure_ino secure[CONFIG_LIDS_MAX_INODE];
struct lids_sys_acl s_acl[CONFIG_LIDS_MAX_SACL];
struct lids_acl o_acl[CONFIG_LIDS_MAX_OACL];
} _lids_data_t;

/***********************************************************************
 ***********************************************************************
 *
 * General variables
 *
 ***********************************************************************
 ***********************************************************************/

int lids_load=0;        /* it is raised to 1 when kernel boot */
int lids_local_on=1;
lids_flags_t lids_flags=0;
int lids_local_pid=0;
unsigned long lids_current=0;

int lids_first_time=1;

kernel_cap_t lids_cap_val=0xffffffff; /* store the capability value in lids.cap */
kernel_cap_t lids_conf_cap=0; /* store the capability value in lids.cap */

#ifdef CONFIG_LIDS_ALLOW_SWITCH
static char lids_pw[160];
#endif

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
static struct allowed_ino lidsadm;
#endif

static _lids_data_t lids_data[2]={{0,0},{0,0}};

#ifdef CONFIG_LIDS_RELOAD_CONF
static int lids_lock_init=0;
static spinlock_t lids_lock;
#endif

extern int lids_read_net(void);

static unsigned long lids_bittab[32]=
{
	0x00000001,0x00000002,0x00000004,0x00000008,
	0x00000010,0x00000020,0x00000040,0x00000080,
	0x00000100,0x00000200,0x00000400,0x00000800,
	0x00001000,0x00002000,0x00004000,0x00008000,
	0x00010000,0x00020000,0x00040000,0x00080000,
	0x00100000,0x00200000,0x00400000,0x00800000,
	0x01000000,0x02000000,0x04000000,0x08000000,
	0x10000000,0x20000000,0x40000000,0x80000000
};

/***********************************************************************
 ***********************************************************************
 *
 * LIDS protection management
 *
 ***********************************************************************
 ***********************************************************************/

int lids_local_off(void) 
{
	struct task_struct *t;
	
	read_lock(&tasklist_lock);
	t=current;
	while (t && (t->pid > 1)) {
		if (t->pid == lids_local_pid) {
			read_unlock(&tasklist_lock);
			return 1;
		}
		t=t->p_pptr;
	}
	read_unlock(&tasklist_lock);
	return 0;
}

/*
 *
 *	LIDS ACL Function 
 *
 */ 
/*
 *	lids_time_checker for file access time restrition
 *
 */ 
static __inline__ int lids_time_checker(const time_t currenttime,time_t time[][2])
{
	time_t from_time,to_time;
	int i;

	for(i=0;i<LIDS_TIME_ITEM&&time[i][0]!=-1;i++) {

		from_time = time[i][0];
		to_time	  = time[i][1];

		/* FIXME: does this really mean if from is 23:00 and to
		   is 01:00 then 01:00<currenttime<23:00 is the valid
		   range? */
		if (from_time > to_time) {
			if (!(currenttime> to_time && currenttime < from_time))
				return 0;
			}
		else if ((currenttime >=from_time && currenttime <= to_time)) {
			return 0;
		}
	}
	/* we may now have an empty time array */
	if (!i)return 0;

	return -1;
}
/*
 *	check the CAP_NET_BIND_SERVICE to bind to specify port
 */ 
int lids_bind_checker(const int port)
{
	int i=0;

	/* if the LIDS is disable , return success */
	if (!(lids_load && lids_local_load)) 
		return 1;
	/* if CAP_NET_BIND_SERVICE is enable global, return success */
	if (cap_raised(lids_cap_val, CAP_NET_BIND_SERVICE))
		return 1;

	/* check only port < 1024) */
	if (port >= 1024 || !current->lids_sys_acl) 
		return 1;

	for(i=0;i<LIDS_PORT_ITEM&&current->lids_sys_acl->port[i][0]!=-1;i++) {
		if (port<=current->lids_sys_acl->port[i][1]&&
				port>=current->lids_sys_acl->port[i][0])
			return 1;
	}

	return 0;
}
/*
 *	lids capability time restrition checker used in capable()
 *	
 */ 
int  lids_cap_time_checker(const int cap)
{
	if (!current->lids_sys_acl) return -1;
	return lids_time_checker(CURRENT_TIME%(60*60*24), current->lids_sys_acl->cap[cap].time);	
	
}
/*
 *	lids capability raising checker
 *
 */
int lids_cap_raised(const int cap)
{
	if ( ( cap_raised(cap_combine(current->lids_cap,lids_cap_val),cap) 
               && !lids_cap_time_checker(cap) )
             || !(lids_load && lids_local_load))
		return 1;
	else
		return 0;
}
/*
 *      used when execed, load the lids_cap into task_struct.
 */

static __inline__ struct lids_sys_acl * lids_do_search_acl(unsigned long int ino,kdev_t dev,_lids_data_t *data)
{
	long i,j;

	/* ok, in case of e.g. 256 entries we would require 256 comparisons
	   for a linear search with no match, the following reduces this
	   to 8 comparisons (table is sorted!) */

	for(j=i=data->search_s_acl;;j>>=1)
	{
		if (i>=data->last_s_acl)i-=j;
		else if (data->s_acl[i].dev<dev)i+=j;
		else if (data->s_acl[i].dev>dev)i-=j;
		else if (data->s_acl[i].ino<ino)i+=j;
		else if (data->s_acl[i].ino>ino)i-=j;
		else return &(data->s_acl[i]);
		if (!j||i<0)return NULL;
	}
}

struct lids_sys_acl * lids_search_acl(unsigned long int ino,kdev_t dev,unsigned long lids_curr)
{
	return lids_do_search_acl(ino,dev,&lids_data[lids_curr&1]);
}

/*
 *	used in before access file(lids_check_base),
 *	check if the program have permission to access the files.
 *	if it has, then dont access the files.
 *
 *	we must sure that current->lids_cap is not NULL when call.
 */ 
static int lids_check_acl(struct dentry *base,int type,unsigned long lids_curr,int check_domain)
{
	struct lids_acl *acl;
	struct dentry *parent;
	struct inode *inode;
	time_t	currenttime;

#ifdef NO_CONFIG_LIDS_RELOAD_CONF
	if (lids_curr!=current->lids_current)
	{
		/* minimal race condition here: in the worst case the
		following code is executed a second time but this
		is better than a global or per process spinlock */
		spin_lock(&lids_lock);

		current->lids_current=lids_current;
		current->lids_sys_acl = lids_do_search_acl(current->lids_ino,current->lids_dev,&lids_data[lids_curr&1]); 
		spin_unlock(&lids_lock);
	}
#endif
	if ( base == NULL || current->lids_sys_acl==NULL)
		return -EPERM; 
	inode  = base->d_inode;
	parent = base->d_parent;
	
	if (parent == NULL)
		return -EPERM;
	
	if ( inode == NULL  ) {
		inode = parent->d_inode;
		parent = parent->d_parent;
	}

	currenttime = CURRENT_TIME;

	do {
		if (check_domain)
			acl = current->lids_sys_acl->lids_domain;
		else
			acl = current->lids_sys_acl->lids_acl;

		while(acl != NULL) {
			if (acl->ino==inode->i_ino && acl->dev==inode->i_dev) {
				return (type & acl->type && !lids_time_checker(currenttime%(60*60*24), acl->time)) ? 0: -EPERM;
			}
			acl = acl->next;
		}
		if ( inode == parent->d_inode) {
			LIDS_DBG("inode == parent->d_inode,return -1\n");
			break;
		}
		inode = parent->d_inode;
	} while( ((parent = parent->d_parent ) != NULL));

	return -1;	
}


int lids_init_task_acl(struct lids_task_acl *acl)
{
	int i;

	acl->lids_cap = 0;
	acl->lids_sys_acl = kmalloc(sizeof(struct lids_sys_acl), GFP_KERNEL);
	if (!acl->lids_sys_acl) {
		LIDS_DBG("kmalloc failed\n");
		return -1;
	}
	memset(acl->lids_sys_acl,0,sizeof(struct lids_sys_acl));
	acl->lids_sys_acl->port[0][0]=-1;
	acl->lids_sys_acl->port[0][1]=-1;
	for(i=0 ; i < 32 ; i++) {
		acl->lids_sys_acl->cap[i].time[0][0]=-1;
		acl->lids_sys_acl->cap[i].time[0][1]=-1;
	}
	return 0;
}

/*
 *	lids_set_acls, this_sys_acl must be NOT NULL.
 */ 
int lids_compute_acls(struct lids_task_acl *current_acl,struct lids_sys_acl *new_sys_acl,struct lids_task_acl *computed_acl)


{

	struct lids_acl *src_acl,*dst_acl;
	int	i,j,k;
	
	/* current->p_pptr,get parent's inherit_flags. 
	 * get the inherit from it's direct parent
	 * if the inherit is not NULL, inherit the capability 
	 * or clear the capability */
	
	LIDS_DBG("compute ACLs for pid %i\n",current->pid);

	if (lids_init_task_acl(computed_acl) < 0) {
		LIDS_DBG("init_task_acl failed\n");
		return -1;
	}

	/* exam the inherit first from current->lids_sys_acl*/
	if (current_acl->lids_sys_acl) {
		/*if (current_acl->lids_sys_acl->forked)*/ {
			LIDS_DBG(" + pid %i: inherit ACLs:\n",current->pid);
			
			/* 1. CAP inherit */
			if (current_acl->lids_sys_acl->flags != 0) {
				computed_acl->lids_sys_acl->flags = 0;
				/* reset the cap_inherit */
				for (i=0 ; i<32 ; i++) {
					if (test_bit(i, &current_acl->lids_sys_acl->flags)) {
						if (current_acl->lids_sys_acl->cap[i].inherit != 0) {
							set_bit(i,&computed_acl->lids_sys_acl->flags);
							memcpy(&computed_acl->lids_sys_acl->cap[i],&current_acl->lids_sys_acl->cap[i],sizeof(struct lids_cap));
							if (current_acl->lids_sys_acl->cap[i].inherit > 0)
								computed_acl->lids_sys_acl->cap[i].inherit--;
							LIDS_DBG("     + pid %i: cap %i inherited. remaining TTL : %i\n",
								 current->pid,i,computed_acl->lids_sys_acl->cap[i].inherit);
						} else {
							LIDS_DBG("     + pid %i: cap %i not inherited: TTL elapsed.\n",current->pid,i);
						}
					}
				}
			}
			LIDS_DBG("     = pid %i: inherit caps %#lx\n",current->pid,computed_acl->lids_sys_acl->flags);
			if (test_bit(CAP_NET_BIND_SERVICE, &computed_acl->lids_sys_acl->flags)) {
				memcpy(computed_acl->lids_sys_acl->port,current_acl->lids_sys_acl->port,sizeof(int)*2*LIDS_PORT_ITEM);
				LIDS_DBG("     + pid %i: port bind acl trasmited\n",current->pid);
			}
			/* 2. File Acls inherit */
			computed_acl->lids_sys_acl->lids_acl=NULL;
			src_acl = current_acl->lids_sys_acl->lids_acl;
			while (src_acl) {
				if (src_acl->inherit != 0) {
					dst_acl= kmalloc(sizeof(struct lids_acl),GFP_KERNEL);			
					if(!dst_acl) {
						LIDS_DBG("kmalloc failed\n");
						lids_free_lids_task_acl(computed_acl);
						return -2;
					}
					memcpy(dst_acl,src_acl,sizeof(struct lids_acl));
					if (dst_acl->inherit > 0) 
						dst_acl->inherit--;
					LIDS_DBG("     + pid %i: 1 ACL inherited. remaining TTL : %i\n",current->pid,dst_acl->inherit);
					dst_acl->next=computed_acl->lids_sys_acl->lids_acl;
					computed_acl->lids_sys_acl->lids_acl=dst_acl=dst_acl;
				} else {
					LIDS_DBG("     + pid %i: 1 ACL not inherited: TTL elapsed.\n",current->pid);
				}
				src_acl=src_acl->next;
			}
			LIDS_DBG("     = pid %i: %s inherit acls\n",current->pid,computed_acl->lids_sys_acl->lids_acl ? "does" : "does not");

/*		computed_acl->lids_sys_acl->forked=0;*/
		
		}
	}

        /* Set its own ACLs and CAP */
	if (new_sys_acl) {
		/* 
		 * FIXME: no merging is done here, neither for time nor for
		 * ports. 
		 */
		
		LIDS_DBG(" + pid %i: getting new ACLs:\n",current->pid);
		computed_acl->lids_sys_acl->flags |= new_sys_acl->flags;

		for (i=0 ; i < 32 ; i++) {
			if (test_bit(i,&new_sys_acl->flags)) {
				/* this may be wrong if both values are negative
				 * but nevertheless this leaves at least 31 bits
				 * to play with. inheritance should be unsigned. 
				 */

				LIDS_DBG("     + pid %i: getting new cap : cap #%i\n",current->pid,i);
				
				/* Here we do an unsigned comparison for -1 to be the biggest number */ 
				if ((unsigned int)computed_acl->lids_sys_acl->cap[i].inherit < (unsigned int)new_sys_acl->cap[i].inherit) {
					computed_acl->lids_sys_acl->cap[i].inherit=new_sys_acl->cap[i].inherit;
					LIDS_DBG("         + pid %i: adjusting cap %i TTL to %i\n",current->pid,i,computed_acl->lids_sys_acl->cap[i].inherit);
				}
				
				for(j=0 ; j < LIDS_TIME_ITEM ; j++) {
					if (computed_acl->lids_sys_acl->cap[i].time[j][0] == -1)
						break;
				}
				for(k=0 ; j < LIDS_TIME_ITEM ; j++,k++) {
					if (new_sys_acl->cap[i].time[k][0]==-1)
						break;
					computed_acl->lids_sys_acl->cap[i].time[j][0]=new_sys_acl->cap[i].time[k][0];
					computed_acl->lids_sys_acl->cap[i].time[j][1]=new_sys_acl->cap[i].time[k][1];
					LIDS_DBG("         + pid %i: adding cap %i time limit %li-%li\n",
						 current->pid,i,new_sys_acl->cap[i].time[k][0],new_sys_acl->cap[i].time[k][1]);
				}
			}
		}
		if (test_bit(CAP_NET_BIND_SERVICE, &new_sys_acl->flags)) {
			LIDS_DBG("     + pid %i: Got CAP_NET_BIND_SERVICE. Adjusting net bind ports\n",current->pid);
			for(j=0; j < LIDS_PORT_ITEM ; j++) {
				if (computed_acl->lids_sys_acl->port[j][0]==-1)
					break;
			}
			
			for(k=0 ; j < LIDS_PORT_ITEM ; j++,k++) {
				computed_acl->lids_sys_acl->port[j][0]=new_sys_acl->port[k][0];
				computed_acl->lids_sys_acl->port[j][1]=new_sys_acl->port[k][1];
				if (new_sys_acl->port[k][0]==-1)
					break;
			}
		}


		/* FIXME:The domain change to current running process's domain */
#if 0
                /* FIXME: either we do domain setup here or while forking. Nevertheless
		 * we may do both but have to do chain handling in this case here.
		 * Otherwise we will leak kernel memory and get unexpected results 
		 */
		computed->lids_sys_acl->lids_domain = new_sys_acl->lids_domain;
#endif

		/* Add ACLs */

		src_acl = new_sys_acl->lids_acl;
		while (src_acl) {
			LIDS_DBG("     + pid %i: getting a new fs ACL\n",current->pid);
			dst_acl= kmalloc(sizeof(struct lids_acl),GFP_KERNEL);			
			if(!dst_acl) {
				LIDS_DBG("kmalloc failed\n");
				lids_free_lids_task_acl(computed_acl);
				return -3;
			}
			memcpy(dst_acl,src_acl,sizeof(struct lids_acl));
			dst_acl->next=computed_acl->lids_sys_acl->lids_acl;
			computed_acl->lids_sys_acl->lids_acl=dst_acl=dst_acl;
			src_acl=src_acl->next;
		}
	}
        computed_acl->lids_cap = computed_acl->lids_sys_acl->flags;

	LIDS_DBG(" = pid %i: final caps : %#lx\n",current->pid,computed_acl->lids_cap);


	return 0;
}




void lids_get_task_acl(struct lids_task_acl *acl,struct task_struct *task)
{
	task_lock(task);
	acl->lids_sys_acl=task->lids_sys_acl;
	acl->lids_cap=task->lids_cap;
#ifdef CONFIG_LIDS_RELOAD_CONF
	acl->lids_current=task->lids_current;
	acl->lids_ino=task->lids_ino;
	acl->lids_dev=task->lids_dev;
#endif
	task_unlock(task);
}	

void lids_set_task_acl(struct lids_task_acl *acl,struct task_struct *task)
{
	task_lock(task);
	task->lids_sys_acl=acl->lids_sys_acl;
	task->lids_cap=acl->lids_cap;
#ifdef CONFIG_LIDS_RELOAD_CONF
	task->lids_current=acl->lids_current;
	task->lids_ino=acl->lids_ino;
	task->lids_dev=acl->lids_dev;
#endif
	task_unlock(task);
}	

int lids_task_acl_deep_copy(struct lids_task_acl *dst,struct lids_task_acl *src)
{
	dst->lids_cap=src->lids_cap;
#ifdef CONFIG_LIDS_RELOAD_CONF
	dst->lids_current=src->lids_current;
	dst->lids_ino=src->lids_ino;
	dst->lids_dev=src->lids_dev;
#endif
	if (src->lids_sys_acl) {
		struct lids_acl *s,*d;
		
		dst->lids_sys_acl=(struct lids_sys_acl *)kmalloc(sizeof(struct lids_sys_acl), GFP_KERNEL);
		if (!dst->lids_sys_acl) {
			LIDS_DBG("kmalloc error\n");
			return -1;
		}
		memcpy(dst->lids_sys_acl,src->lids_sys_acl, sizeof(struct lids_sys_acl));

		/* 1 . copy lids_acl */
		dst->lids_sys_acl->lids_acl=NULL;
		s=src->lids_sys_acl->lids_acl;
		while (s) {
			d = kmalloc(sizeof(struct lids_acl),GFP_KERNEL);
			if (!d) {
				LIDS_DBG("kmalloc error\n");
				lids_free_lids_task_acl(dst);
				return -1;
			}
			memcpy(d,s,sizeof(struct lids_acl));
			d->next = dst->lids_sys_acl->lids_acl;
			dst->lids_sys_acl->lids_acl = d;
			s = s->next;
		}

		/* 2. copy lids_domain */
		dst->lids_sys_acl->lids_domain=NULL;
		s=src->lids_sys_acl->lids_domain;
		while (s) {
			d = kmalloc(sizeof(struct lids_acl),GFP_KERNEL);
			if (!d) {
				LIDS_DBG("kmalloc error\n");
				lids_free_lids_task_acl(dst);
				return -1;
			}
			memcpy(d,s,sizeof(struct lids_acl));
			d->next = dst->lids_sys_acl->lids_domain;
			dst->lids_sys_acl->lids_domain = d;
			s = s->next;
		}
	} else
		dst->lids_sys_acl=NULL;
	return 0;
}


void lids_free_lids_task_acl(struct lids_task_acl *acl)
{
        struct lids_sys_acl * this_sys_acl = acl->lids_sys_acl;
        struct lids_acl *this_acl,*next_acl;

        if (this_sys_acl) {
		this_acl = this_sys_acl->lids_acl;
                while(this_acl) {
			next_acl = this_acl->next;
                      	kfree(this_acl);
			this_acl = next_acl;
                }
		this_acl = this_sys_acl->lids_domain;
                while(this_acl) {
			next_acl = this_acl->next;
                      	kfree(this_acl);
			this_acl = next_acl;
                }
                LIDS_DBG("pid=%d begin to free %p\n",current->pid,this_sys_acl);
    		kfree(this_sys_acl);
		acl->lids_sys_acl=NULL;
        }
}	

/*
 *      free allocated current->lids_sys_acl when do_exit()
 */
void exit_lids(struct task_struct *tsk)
{
	struct lids_task_acl acl;

	lids_get_task_acl(&acl,tsk);
	tsk->lids_sys_acl=NULL;
	lids_free_lids_task_acl(&acl);
}

/*
 *	lids_add_acl 
 *	add the acl into system.
 */ 
static void lids_add_acl(unsigned long int s_ino ,kdev_t s_dev , unsigned long int o_ino,kdev_t o_dev, int type, int inherit, _lids_data_t *data, void *time, void *port) 
{
        struct lids_sys_acl *this_sys_acl;
        struct lids_acl *lids_acl,*acl;
	int i;

	for(this_sys_acl=NULL,i=0;i<data->last_s_acl&&!this_sys_acl;i++)
	    if (data->s_acl[i].ino==s_ino&&data->s_acl[i].dev==s_dev)
		this_sys_acl=&(data->s_acl[i]);

        if (!this_sys_acl)
	{
		if (data->last_s_acl==CONFIG_LIDS_MAX_SACL)return;
		if ((int)(o_ino)!=-1&&data->last_o_acl==CONFIG_LIDS_MAX_OACL)
			return;

	        LIDS_DBG("lids_add_s_acl,%ld,%d\n",s_ino,s_dev);
 
        	this_sys_acl=&(data->s_acl[(data->last_s_acl)++]);
	        this_sys_acl->ino = s_ino;
	        this_sys_acl->dev = s_dev;
		this_sys_acl->flags = 0;
		memset(this_sys_acl->cap,'\0',32*sizeof(struct lids_cap *));
		this_sys_acl->lids_acl = NULL;
		this_sys_acl->lids_domain = NULL;
		this_sys_acl->port[0][0]=this_sys_acl->port[0][1]=-1;
	}
	/* if it is a capabilities */
	if (type == LIDS_CAP) 
	{
		/* flags indicate that this proceed has capability 
		 * to do someting ,inherit_flags dicide that if the capability
		 * can be inherit to its children */
		set_bit(o_dev,&this_sys_acl->flags);
		this_sys_acl->cap[o_dev].inherit = inherit;
		
		memcpy(this_sys_acl->cap[o_dev].time,time,
			sizeof(this_sys_acl->cap[o_dev].time));

		if (o_dev == CAP_NET_BIND_SERVICE) 
			memcpy(this_sys_acl->port,port,
				sizeof(this_sys_acl->port));
		return;
	}

	if (data->last_o_acl==CONFIG_LIDS_MAX_OACL)return;

        LIDS_DBG("lids_add_o_acl,%ld,%d,%ld,%d,%d\n",this_sys_acl->ino,this_sys_acl->dev,o_ino,o_dev,type);

	lids_acl=&(data->o_acl[(data->last_o_acl)++]);
	lids_acl->ino = o_ino;
	lids_acl->dev = o_dev;
	lids_acl->inherit = inherit;
	memcpy(lids_acl->time,time,LIDS_TIME_ITEM*2*sizeof(time_t*));
	lids_acl->next = NULL;
	
        if (type < 0) {
                lids_acl->type = -type;
                if ((acl=this_sys_acl->lids_domain)==NULL)
                        this_sys_acl->lids_domain=lids_acl;
                else {
                        while(acl->next)acl=acl->next;
                        acl->next = lids_acl;
                }
        }
        else {
                lids_acl->type = type;
                if ((acl=this_sys_acl->lids_acl)==NULL)
                        this_sys_acl->lids_acl=lids_acl;
                else {
                        while(acl->next)acl=acl->next;
                        acl->next = lids_acl;
                }
        }
}

static __inline__ int lids_acl_dev_ino_cmp(struct lids_sys_acl *s1,struct lids_sys_acl *s2)
{
	if (s1->dev<s2->dev)return -1;
	if (s1->dev>s2->dev)return 1;
	if (s1->ino<s2->ino)return -1;
	if (s1->ino>s2->ino)return 1;
	return 0;	/* Aieeeh, duplicate entry */
}

static int lids_sort_acls(_lids_data_t *data)
{
	int i;
	int working;
	struct lids_sys_acl tmp;

	/* sort source acl entries by device and inode */

	do
	{
	    working=0;
	    for(i=0;i<data->last_s_acl-1;i++)
	      switch(lids_acl_dev_ino_cmp(&(data->s_acl[i]),
		&(data->s_acl[i+1])))
	    {
		case 0:	return 0;
		case 1: tmp=data->s_acl[i];
			data->s_acl[i]=data->s_acl[i+1];
			data->s_acl[i+1]=tmp;
			working=1;
	    }
	    for(i=data->last_s_acl-1;i>0;i--)
	      switch(lids_acl_dev_ino_cmp(&(data->s_acl[i-1]),
		&(data->s_acl[i])))
	    {
		case 0:	return 0;
		case 1: tmp=data->s_acl[i-1];
			data->s_acl[i-1]=data->s_acl[i];
			data->s_acl[i]=tmp;
			working=1;
	    }
	} while(working);

	/* set up base index for fast search: no, we can't use >0x20000000
	   here but anyway it'll take some years for computers to
	   allow for the memory required for this size :-) */

	data->search_s_acl=0x20000000;
	while(data->last_s_acl<=data->search_s_acl&&data->search_s_acl)
		data->search_s_acl>>=1;

	return 1;
}

/***********************************************************************
 *
 *	lids_search_inode()	
 *
 *	check the inode in a arr_ino array.
 *	may be use fast search in this array.
 *	xhg, 1999/10/11.
 *
 */

static __inline__ int lids_search_inode(unsigned long ino,kdev_t dev,_lids_data_t *data) 
{
	long i=(ino^dev)&0xffff;
	long j;

	/* when the 'hash' bit is not set we for sure don't have
	   a matching entry, so just signal 'nothing found' */

	if (!(data->fastguess[i>>5]&lids_bittab[i&31]))return -1;

	/* ok, in case of e.g. 256 entries we would require 256 comparisons
	   for a linear search with no match, the following reduces this
	   to 8 comparisons (table is sorted!) */

	for(j=i=data->search_secure;;j>>=1)
	{
		if (i>=data->last_secure)i-=j;
		else if (data->secure[i].dev<dev)i+=j;
		else if (data->secure[i].dev>dev)i-=j;
		else if (data->secure[i].ino<ino)i+=j;
		else if (data->secure[i].ino>ino)i-=j;
		else
		{
			return data->secure[i].type;
		}
		if (!j||i<0)return -1;
	}
}

int lids_check_hidden_inode(unsigned long ino,kdev_t dev)
{
	if (lids_search_inode(ino,dev,&lids_data[lids_current&1])==LIDS_DENY)
		return -1;
	return 0;
}

/***********************************************************************
 *
 *	lids_check_base();
 *	
 *	check if the base have been protected by the IDS system.
 *	use the base->d_parent 
 *	check if the requried access can be permitted
 */ 

int lids_check_base(struct dentry *base, int flag)
{
	struct inode *ino;
	struct dentry *dentry=base;
	int retval=0;
	unsigned long lids_curr=lids_current;

        /* check if the dentry is in the domain */
        while(dentry) {
                if ((ino=dentry->d_inode)!=NULL) {
                        if (current->lids_sys_acl && current->lids_sys_acl->lids_domain)
                                retval = lids_check_acl(base,flag,lids_curr,1);
                        else
                                break;
                        if (retval == 0)
                                break;
                        return -EPERM;
                }
                if (dentry==dentry->d_parent) {
                        /* do not in this exec domain */
                        return -EPERM;
                }
                dentry=dentry->d_parent;
        }
	
	while (dentry) {
		if ((ino=dentry->d_inode)!=NULL)
		    if ((retval=lids_search_inode(ino->i_ino,ino->i_dev,
			&lids_data[lids_curr&1])) >= 0) {
			return (retval & flag) ? 0:
				lids_check_acl(base,flag,lids_curr,0);
		}
		if (dentry==dentry->d_parent)
			return 0;
		dentry=dentry->d_parent;
	}

	return 0;
}

/***********************************************************************
 *
 *	lids_add_inode(inode,dev,type)
 *
 *	add the given inode into the arr_ino.
 *	if full return 0
 *	else return the insert position.
 *
 */
static void lids_add_inode(unsigned long inode ,kdev_t dev , int type,_lids_data_t *data) 
{
	int i;

	if (data->last_secure==CONFIG_LIDS_MAX_INODE)return;

	LIDS_DBG("lids_add_inode,%ld,%d,%d\n",inode,dev,type);

	for (i=0;i<data->last_secure;i++)
	    if (data->secure[i].ino==inode&&data->secure[i].dev==dev)
		return;

	data->secure[data->last_secure].ino = inode;
	data->secure[data->last_secure].dev = dev;
	data->secure[data->last_secure].type = type;

	(data->last_secure)++;

	i=(dev^inode)&0xffff;
	data->fastguess[i>>5]|=lids_bittab[i&31];
}

static __inline__ int lids_inode_dev_ino_cmp(struct secure_ino *s1,struct secure_ino *s2)
{
	if (s1->dev<s2->dev)return -1;
	if (s1->dev>s2->dev)return 1;
	if (s1->ino<s2->ino)return -1;
	if (s1->ino>s2->ino)return 1;
	return 0;	/* Aieeeh, duplicate entry */
}

static int lids_sort_inodes(_lids_data_t *data)
{
	int i;
	int working;
	struct secure_ino tmp;

	/* sort inode entries by device and inode */

	do {
	    working=0;
	    for(i=0;i<data->last_secure-1;i++)
	      switch(lids_inode_dev_ino_cmp(&(data->secure[i]),
		&(data->secure[i+1]))) {
		case 0:	return 0;
		case 1: tmp=data->secure[i];
			data->secure[i]=data->secure[i+1];
			data->secure[i+1]=tmp;
			working=1;
	      } 
	    for(i=data->last_secure-1;i>0;i--)
	      switch(lids_inode_dev_ino_cmp(&(data->secure[i-1]),
		&(data->secure[i]))) {
		case 0:	return 0;
		case 1: tmp=data->secure[i-1];
			data->secure[i-1]=data->secure[i];
			data->secure[i]=tmp;
			working=1;
	      }
	} while(working);

	/* set up base index for fast search: no, we can't use >0x20000000
	   here but anyway it'll take some years for computers to
	   allow for the memory required for this size :-) */

	data->search_secure=0x20000000;
	while(data->last_secure<=data->search_secure&&data->search_secure)
		data->search_secure>>=1;

	return 1;
}

/**************************************************************************/

/*
 *	lids_get_info(char *buffer);
 *
 */ 
static int lids_get_info(char *buffer, unsigned long int *ino,kdev_t *dev,int type)
{
	int	error = -1;
	char  	ino_str[64],dev_str[64];
	char 	*q,*p;
	struct 	dentry *d_file;
	struct nameidata nd;

	/* get the dentry of the file */
	/* it is a special type, ino=-1*/

	memset(ino_str,'\0',64);
	memset(dev_str,'\0',64);

	p = memscan(buffer,':',strlen(buffer));
	if (((unsigned long)(p-buffer))==strlen(buffer))goto exit;
	if (((unsigned long)(p-buffer))>63)goto exit;
	memcpy(ino_str,buffer,p-buffer);

	p++;
	q = memscan(p,':',strlen(p));
	if (((unsigned long)(q-p))==strlen(p))goto exit;
	if (((unsigned long)(q-p))>63)goto exit;
	memcpy(dev_str,p,q-p);

	error = 0;
	if (type == LIDS_CAP) {
		*dev = (kdev_t)simple_strtoul(dev_str,0,0);
		*ino = -1;
		goto exit;
	}
	q++;
	error = user_path_walk(q,&nd);
	d_file = nd.dentry;

	if (error) {
		*ino = simple_strtoul(ino_str,0,0);
		*dev = (kdev_t)simple_strtoul(dev_str,0,0);
	}
	else {
		if (d_file) if (d_file->d_inode){
		if ( (d_mountpoint(d_file))) { 
			LIDS_DBG("OK,Mount point found!\n");
			*ino = simple_strtoul(ino_str,0,0);
			*dev = (kdev_t)simple_strtoul(dev_str,0,0);
		}
		else {
			*ino=simple_strtoul(ino_str,0,0);
			*dev=(kdev_t)simple_strtoul(dev_str,0,0) ;
			if ((*ino != d_file->d_inode->i_ino) ||
			    (*dev != d_file->d_inode->i_dev)) {
				LIDS_DBG("now=%p\n",d_file);
				printk("LIDS: [%s] dev/inode in lids.conf seems wrong\n",q);
			}
		}
		}
		path_release(&nd);
	}
	error = 0;
 exit:	
	return error;
}
/*
 *	lids_str2data
 *
 */ 
int lids_str2data(char *string, int data[][2],int item)
{
	
	char *p=NULL;
	char *q=NULL;
	char *r=NULL;
	int  i=0;

	r = string;
	p = strchr(r,':');
	if (p!=NULL) *p = '\0';
	do {
		q = strchr(r,',');
		if (q != NULL) *q = '\0';

		p = strchr(r,'-');
		if ( p == NULL ) return -1;
		*p = '\0';

		data[i][0] = simple_strtoul(r,0,0);
		data[i][1] = simple_strtoul(p+1,0,0);
		/* big bad problem later if we do allow 0-0 values to
		   be stored here: merging of time/port data will fail */
		if (data[i][0]||data[i][1])i++;
		if (q!=NULL) r = q+1;
	}while( q != NULL && i< item );
	/* marked the lastest item */
	if (i!=item) data[i][0]=data[i][1]=-1;
	return 0;
}

/***********************************************************************
 *
 *	lids_init_add_file ( char * filename)
 *
 *	read filename from read_file() 
 *
 *	method : get the dentry from the filename.
 *		 check if its parent is in the err_ino.
 *		 if not, add the inode.
 *		 
 *	the format is s_ino:s_dev:s_file:type:o_ino:o_dev:o_file
 */

static int lids_init_add_file ( char *buffer,_lids_data_t *data)
{
	char *p,*q;
	int error = -1;
	int is_default_rule = 0;
	int type,inherit;
	unsigned long int s_ino,o_ino;
	kdev_t	s_dev,o_dev;
	int	time[LIDS_TIME_ITEM][2];
	int	port[LIDS_PORT_ITEM][2];

	p = memscan(buffer,':',strlen(buffer));
	if (((unsigned long)(p-buffer))==strlen(buffer))goto exit;

	p++;
	q = memscan(p,':',strlen(p));
	error = -2;
	if (((unsigned long)(q-p))==strlen(p))goto exit;

	error = -3;
	q++;
	if (*q == ':' ) {
		is_default_rule = 1;
		p = q;
	}
	else {
		p = memscan(q,':',strlen(q));
		if (((unsigned long)(p-q))==strlen(q))goto exit;
	}
        *p = '\0';
	p++;
	q = memscan(p,':',strlen(p));
	*q = '\0';
	q++;
	type =simple_strtol(p,0,0);

	p = memscan(q,':',strlen(q));
	*p = '\0';
	inherit = simple_strtol(q,0,0);
	p++;

	q = strrchr(p,':');
	error = -3;
	if (q == NULL)
		goto exit;
	*q = '\0';
	q++;

	/* p point to the object's ino,dev,filename */
	error = -4;
	if (is_default_rule) {
		s_ino = s_dev = 0;
	}	
	else {
		if ( lids_get_info(buffer,&s_ino,&s_dev,0) < 0 )
			goto exit;
	}
	error = -5;
	/* get the object */
	if ( lids_get_info(p,&o_ino,&o_dev,type) < 0 )
		goto exit ;

	/* get the port scale */
	error = -6;
	if (type == LIDS_CAP && o_dev == CAP_NET_BIND_SERVICE) 
/* FIXME: no guarantee that time_t and int are of the same size, this
   needs to be refined to data size independence */
		if ( lids_str2data(p,port,LIDS_PORT_ITEM) < 0) 
			goto exit;
	/* get the time here */
	error = -7;
	if (lids_str2data(q,time,LIDS_TIME_ITEM) < 0) 
		goto exit;
	if (is_default_rule)
		lids_add_inode(o_ino,o_dev,type,data);
	else {
		lids_add_acl(s_ino,s_dev,o_ino,o_dev,type,inherit,data,time,port);
	}
	error = 0;
exit:
	return error;
}

/***********************************************************************
 *
 *	add a file into the secure list.
 *
 */

#if 0  /* not used */
static int init_add_file(char *filename,_lids_data_t *data) 
{
	struct dentry *dentry;
	int	error=0;
	struct nameidata nd;

	if (path_init(filename, LOOKUP_FOLLOW|LOOKUP_POSITIVE, &nd))
		error = path_walk(filename, &nd);
	if (error)
		goto err_out;
	dentry = nd.dentry;
     	
	if (IS_ERR(dentry) || !dentry || !(dentry->d_inode)) 
		error = -1;
	else { 
		lids_add_inode(dentry->d_inode->i_ino,dentry->d_inode->i_dev,
				LIDS_DENY,data);
	}
	path_release(&nd);
err_out:
	LIDS_DBG("init_add_file : return %d\n",error);
	return error;
}
#endif 

/*
 * read the capability value into lids_cap 
 * format [+|-]number:cap_name
 */
int lids_cap_init(char *buffer)
{
	char *p;
	int flag;
	int error = 0;
	error = -1;	
	p = memscan(buffer,':',strlen(buffer));
	if (((unsigned long)(p-buffer))==strlen(buffer))goto exit;
	
	*p = '\0';
	flag = simple_strtoul(buffer+1,0,0);
	error = 0;
	if (buffer[0] == '+')
		cap_raise(lids_conf_cap,flag);
	else if (buffer[0] == '-' )
	       	cap_lower(lids_conf_cap,flag);
	else 	
		error = -2;
exit:
	return error;
}

/*
 * lids read capability from /etc/lids/lids.cap 
 */

int lids_read_cap(void)
{
 	struct file     *filp;
        char    buffer[1024], *p,*q;
        mm_segment_t    oldfs;
        int     bytes;
	int 	error =0;
	int 	start=0,finished=0;

        filp = filp_open(LIDS_CAP_FILE,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -1 ;
		printk("LIDS: Error opening capability file " LIDS_CAP_FILE ". Does it exist?\n");
		/* FIXME: if (lids_load) goto err_panic;  */
		return error;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file " LIDS_CAP_FILE " can not be read\n");
		/*
            	if (lids_load) goto err_panic ; 
		*/
		return error;
	}
	while ( !finished ) {
       		filp->f_pos = start;
       		oldfs = get_fs();
       		set_fs(KERNEL_DS);
        	bytes = filp->f_op->read(filp,buffer,1024,&filp->f_pos);
        	set_fs(oldfs);

		q=buffer;

		if (bytes<1024){
			finished=1;
			q[bytes++]='\n';
		}
      /* Now read 1024 bytes into buffer */
       /* analyze the buffer */
		while(bytes&&(p=memscan(q,'\n',bytes))!=q+bytes) {
			*p++='\0';
			start+=(int)(p-q);
			bytes-=(int)(p-q);
			while(*q=='\r')q++; /* hmmm... */
			if (*q=='-' || *q=='+') if ( lids_cap_init(q) < 0) {
				error = -5;
				printk("LIDS: error in adding [%s] to the kernel\n",q);
				break;
			}
			q=p;
		}
		if (error)break;

		if (bytes==1024)
		{
		    printk("LIDS: Line too long in file %s\n",LIDS_CAP_FILE);
		    error = -4;	/* if the line contains no '\n' */
		    break;
		}
	}
	LIDS_DBG("lids_read_cap:lids_conf_cap = %x\n",lids_conf_cap);
       /* Close the file */
       	fput(filp);
	return error;	
}
/*
 *  read the lids password generated by "lidsadm -P" into kernel 
 *  if nessary 
 */ 
#ifdef CONFIG_LIDS_ALLOW_SWITCH
int lids_read_pw(void)
{
 	struct file     *filp;
        char    buffer[LIDS_PW_LEN];
        mm_segment_t    oldfs;
        int     bytes;
	int 	error =0;

        filp = filp_open(LIDS_PW_FILE,O_RDONLY,O_RDONLY);
        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -1 ;
		printk("LIDS: Error opening passwd file " LIDS_PW_FILE ". Does it exist?\n");
		return error;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file " LIDS_PW_FILE " can not be read\n");
		return error;
	}

      /* Now read LIDS_PW_LEN bytes from postion "StartPos" */
       filp->f_pos = 0;
       oldfs = get_fs();
       set_fs(KERNEL_DS);
       bytes = filp->f_op->read(filp,buffer,LIDS_PW_LEN,&filp->f_pos);
       set_fs(oldfs);

	if (bytes<LIDS_PW_LEN)
		return -1;

	memcpy(lids_pw,buffer,LIDS_PW_LEN);

       /* Close the file */
        fput(filp);
	return error;	
}
#endif

/***********************************************************************
 *
 *	lids_init
 *
 *	initialize the vfs security system. read the config file .
 *	add the inode to the files.
 *
 */

int lids_init(void)
{
 	struct file     *filp;
        char    buffer[1024], *p,*q;
        mm_segment_t    oldfs;
        int     bytes;
	int	start = 0;
	int	finished=0;
	int	error = 0;
	_lids_data_t *data=&lids_data[(lids_current&1)^1];
#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
	struct dentry *dentry;
	struct nameidata nd;
#endif


	/* Get lidsadm dev/inode */
	
	LIDS_DBG("into lids_init_..\n");
#ifdef CONFIG_LIDS_RELOAD_CONF
	if (!lids_lock_init)
	{
		spin_lock_init(&lids_lock);
		lids_lock_init=1;
	}
#endif

	lids_cap_val = cap_bset;

	/* reset fast guess table */

	memset(data->fastguess,0,sizeof(data->fastguess));

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
	if (path_init(LIDS_ADM_PATH, LOOKUP_FOLLOW|LOOKUP_POSITIVE,&nd))
		error = path_walk(LIDS_ADM_PATH,&nd);
	if (error)
		return -1;
	/*
	down(&nd.dentry->d_inode->i_sem);
	dentry = lookup_hash(&nd.last, nd.dentry);
 	*/
	dentry = nd.dentry;

  	if (IS_ERR(dentry) || (!dentry)) {
    		printk("LIDS : " LIDS_ADM_PATH " not found \n");
		lidsadm.ino = 0;
		lidsadm.dev = 0;
  	}
	else {
		if (!dentry->d_inode ||
		   !dentry->d_inode->i_dev ||
		   !dentry->d_inode->i_ino) {
			printk("LIDS: something wrong with " LIDS_ADM_PATH "\n");
			lidsadm.ino = 0;
			lidsadm.dev = 0;
		}
		else {
			lidsadm.ino = dentry->d_inode->i_ino;
			lidsadm.dev = dentry->d_inode->i_dev;
		}
	}
	path_release(&nd);
#endif
	/* Now, try to read lids.conf */

/*	if ( init_add_file(LIDS_CONF_DIR,data) < 0 ) {
		error = -1;
		printk("LIDS: Adding conf file %s error,does it exist?\n",LIDS_CONF_FILE);
		if (lids_load) goto lids_panic;
		return -1;
	}
*/

	
        filp = filp_open(LIDS_CONF_FILE,O_RDONLY,O_RDONLY);

        if (IS_ERR(filp)||(filp==NULL)) {
	   	error = -2 ;
		printk("LIDS: Error opening the config file " LIDS_CONF_FILE ". Does it exist?\n");
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}

        if (filp->f_op->read==NULL) {
		fput(filp);
	    	error = -3 ;
		printk("LIDS: The file %s can not be read\n",LIDS_CONF_FILE);
            	if (lids_load) goto lids_panic ;  /* File(system) doesn't allow reads */
		return -1;
	}

      /* Now read 1024 bytes from postion "StartPos" */
	while ( !finished ) {
       		filp->f_pos = start;
       		oldfs = get_fs();
       		set_fs(KERNEL_DS);
        	bytes = filp->f_op->read(filp,buffer,1024,&filp->f_pos);
        	set_fs(oldfs);

		q=buffer;

		if (bytes<1024) {
			finished=1;
			q[bytes++]='\n';
		}

		while(bytes&&(p=memscan(q,'\n',bytes))!=q+bytes) {
		    *p++='\0';
		    LIDS_DBG(" read_file : line=->%s<-\n",q);
		    start+=(int)(p-q);
		    bytes-=(int)(p-q);
		    while (*q == '\r') q++; /* hmmm... */
		    if (*q && (*q !='#'))
			    if (lids_init_add_file(q,data) < 0) {
				    printk("LIDS: error in adding [%s] to the kernel\n",q);
				    error = -5;
				    break;
			    }
		    q=p;
		}

		if (error) break;

		if (bytes == 1024) {
		    printk("LIDS: Line too long in file %s\n",LIDS_CONF_FILE);
		    error = -4;	/* if the line contains no '\n' */
		    break;
		}
	}

	memset(buffer,'\0',1024);

       /* Close the file */
        fput(filp);

	if (error)
	{
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}
	if (!lids_sort_inodes(data))
	{
		printk("LIDS: Aieeh, duplicate inode entry in file %s\n",LIDS_CONF_FILE);
		error = -6;
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}

	if (!lids_sort_acls(data))
	{
		printk("LIDS: Aieeh, duplicate source acl entry in file %s\n",LIDS_CONF_FILE);
		error = -7;
           	if (lids_load) goto lids_panic;  /* Or do something else */
		return -1;
	}
#ifdef CONFIG_LIDS_ALLOW_SWITCH
	/* Read the password now */
	if (lids_read_pw()) {
		printk("LIDS: Read password file error\n");
		error = -8;
		goto lids_panic;
	}
#endif
	/* read the cap init from lids.cap into lids_conf_cap */
	if (lids_read_cap()) {
		printk("LIDS: Read capability file error\n");
		error = -9;
		goto lids_panic;
	}
#ifdef CONFIG_LIDS_SA_THROUGH_NET
	/* read network parameter from lids.net */
	if (lids_read_net()) {
		printk("LIDS: Read mail parameter error\n");
		error = -10;
		goto lids_panic;
	}
#endif

#if 0 
	printk("read_file : finish=%d ret %d\n",finished,error);
	for(start=0;start<data->last_secure;start++)
	{
		struct lids_sys_acl *this_sys_acl;
		struct lids_acl *this_acl;
		int i;

		this_sys_acl = (struct lids_sys_acl *)&(data->s_acl[start]);

		printk("-----------\ndefault: %ld--%d--%d-\n",data->secure[start].ino, data->secure[start].dev, data->secure[start].type);
		for(i=0;i<32;i++) 
			LIDS_DBG("CAPABILITY [%d]:%d,TIME %ld-%ld\t",i,this_sys_acl->cap[i].inherit,this_sys_acl->cap[i].time[0][0],this_sys_acl->cap[i].time[0][1]);

		this_acl = this_sys_acl->lids_acl;
		while(this_acl!=NULL) {
			LIDS_DBG("----------\nobject:\tthis_acl=%ld,%d,type=%d,inherit=%d,from %ld to %ld\n",this_acl->ino,this_acl->dev,this_acl->type,this_acl->inherit,this_acl->from_time,this_acl->to_time);
			this_acl = this_acl->next;
		}

		this_acl = this_sys_acl->lids_domain;
		while(this_acl!=NULL) {
			printk("object:\tthis_domain=%ld,%d,type=%d\n",this_acl->ino,this_acl->dev,this_acl->type);
			this_acl = this_acl->next;
		}
	}
#endif
	if (!error ) { 
		LIDS_DBG("written to index %ld\n",(lids_current&1)^1);
		printk("LIDS: Statistics: %d objects, %d source ACLS, %d object ACLs,capability = %x\n",data->last_secure,data->last_s_acl,data->last_o_acl,lids_conf_cap);
		lids_current++;
		LIDS_DBG("lids_current=%ld index=%ld\n",lids_current,lids_current&1);
		return 0;
	}
	else 
lids_panic:
	panic("LIDS: Cannot initialize the lids system, return code %d",error);
	return error;
}

/***********************************************************************
 ***********************************************************************
 *
 * Now, the sysctl procedures
 *
 ***********************************************************************
 ***********************************************************************/


/***********************************************************************
 *
 * The one which process flags changes
 *
 * If it returns 0, caps won't be processed.
 *
 */


static int lids_process_flags(lids_flags_t flags)
{
#ifdef CONFIG_LIDS_RELOAD_CONF
	int old_lids_local_on;
	int old_lids_local_pid;
#endif
	_lids_data_t *data=&lids_data[(lids_current&1)^1];
	LIDS_DBG("Process flags %#0x\n",flags);
	LIDS_DBG("lids_flag_raised(flags,LIDS_FLAGS_LIDS_ON))=%i\n",lids_flag_raised(flags,LIDS_FLAGS_LIDS_ON));
	LIDS_DBG("lids_flag_raised(flags,LIDS_FLAGS_LIDS_LOCAL_ON)=%i\n",lids_flag_raised(flags,LIDS_FLAGS_LIDS_LOCAL_ON));
	LIDS_DBG("lids_flag_raised(flags,LIDS_FLAGS_RELOAD_CONF)=%i\n",lids_flag_raised(flags,LIDS_FLAGS_RELOAD_CONF));
#ifdef CONFIG_LIDS_ALLOW_SWITCH
		
	/* LIDS switching */
	if ( lids_load != (lids_flag_raised(flags,LIDS_FLAGS_LIDS_ON) != 0) ) {
		lids_load=(lids_flag_raised(flags,LIDS_FLAGS_LIDS_ON) != 0);
		lids_log(0,"LIDS switched to %i",lids_load);
		if (lids_load)
			lids_flag_raise(lids_flags,LIDS_FLAGS_LIDS_ON);
		else
			lids_flag_lower(lids_flags,LIDS_FLAGS_LIDS_ON);
	}
	if ( lids_local_on != (lids_flag_raised(flags,LIDS_FLAGS_LIDS_LOCAL_ON) != 0) ) {
		lids_local_on=(lids_flag_raised(flags,LIDS_FLAGS_LIDS_LOCAL_ON) != 0);   /* XXX: Race condition here. We must first assign the PID */
		lids_log(0,"LIDS locally switched to %i",lids_local_on);
		if (lids_local_on) {
			lids_flag_raise(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON);
		}
		else {
			lids_local_pid=current->p_pptr->pid;
			if (lids_local_pid == 1) {  /* this doesn't apply to init */
				printk("Can't give local lids deactivation to init!!\n");
				lids_flag_raise(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON);
				lids_local_on = 1;
			}
			else
				lids_flag_lower(lids_flags,LIDS_FLAGS_LIDS_LOCAL_ON); /* should not be necessary */			
		}
	}
#else
	if  (!lids_flag_raised(flags,LIDS_FLAGS_LIDS_ON)) {
		lids_security_alert("Attempt to switch LIDS off (feature disabled)");
		return 0;
	}
#endif

	/* Config file reload */
	if (lids_flag_raised(flags,LIDS_FLAGS_RELOAD_CONF)) {
#ifdef CONFIG_LIDS_RELOAD_CONF
		if (lids_load && !lids_local_on && lids_local_load)  {
			printk(KERN_WARNING "Can't reload config files if an LFS is opened and we are not in\n");
		}
		else {
			old_lids_local_on=lids_local_on;
			old_lids_local_pid=lids_local_pid;
			if (lids_load && lids_local_load) {
				LIDS_DBG("Let's give lidsadm (pid %i) the right to read the conf\n",current->pid);
				lids_local_pid=current->pid;
				lids_local_on=0;
			}
			data->last_secure=0;
			data->last_s_acl=0;
			data->last_o_acl=0;
			lids_init();
			lids_security_alert("Config. file reloaded");
			lids_local_pid=old_lids_local_pid;
			lids_local_on=old_lids_local_on;
		}
			
#else
		lids_security_alert("Attempt to reload config. file (feature disabled)");
#endif 
	}
	return 1;
}


/***********************************************************************
 *
 * The one which set/get security features
 *
 */

static int number_failed=0;
static int wait_after_fail=0;
#ifdef CONFIG_LIDS_ALLOW_SWITCH
static struct timer_list fail_timer;
#endif

/* called by timer */
static void reenable_sysctl(unsigned long user_data) 
{
        number_failed=0;
        wait_after_fail=0;
}





int lids_proc_locks_sysctl(ctl_table *table, int write, struct file *filp,
			    void *buffer, size_t *lenp) 
{
	lids_locks_t locks;
	char rmd160sig[LIDS_PW_LEN+10];

#ifdef CONFIG_LIDS_ALLOW_SWITCH
	byte   hashcode[RMDsize/8];
	int i;
#endif
	
	/* first: check the terminal and the program which access the sysctl */

#ifdef CONFIG_LIDS_RESTRICT_MODE_SWITCH
	if (current->tty
	    && ! (0
#ifdef CONFIG_LIDS_MODE_SWITCH_CONSOLE
		  || current->tty->driver.type == TTY_DRIVER_TYPE_CONSOLE
#endif
#ifdef CONFIG_LIDS_MODE_SWITCH_SERIAL
		  || current->tty->driver.type == TTY_DRIVER_TYPE_SERIAL
#endif
#ifdef CONFIG_LIDS_MODE_SWITCH_PTY
		  || current->tty->driver.type == TTY_DRIVER_TYPE_PTY
#endif
		  )) {
		lids_security_alert("Attempt to %s locks sysctl (unauthorized terminal)",
				    write ? "write" : "read");
		return -EPERM;
	}
#endif /* CONFIG_LIDS_RESTRICT_MODE_SWITCH */

#ifndef CONFIG_LIDS_ALLOW_ANY_PROG_SWITCH
	{
		struct vm_area_struct * vma;
		unsigned long ino;
		dev_t dev;
		
		ino=-1;
		dev=-1;
		if (current->mm) {
			vma = current->mm->mmap;
			while (vma) {
				if ((vma->vm_flags & VM_EXECUTABLE) && 
				    vma->vm_file && vma->vm_file->f_dentry && 
				    vma->vm_file->f_dentry->d_inode) {
					ino=vma->vm_file->f_dentry->d_inode->i_ino;
					dev=vma->vm_file->f_dentry->d_inode->i_dev;
					break;
				}
				vma = vma->vm_next;
			}
		}
		if ( (ino != lidsadm.ino) || (dev != lidsadm.dev) ) {
			lids_security_alert("Attempt to %s locks sysctl (unauthorised program)",
					    write ? "write" : "read");
			return -EPERM;
		}
	}
#endif
	/* second: check wether it is not a timeout period after two many failed attempts */

	if (wait_after_fail) {
		lids_security_alert("Attempt to %s locks sysctl during timeout",write ? "write" : "read");
		return -EPERM;
	}
	
	if (write) {
		/* Third : check what is submitted (size, magics, passwd) */
		if (*lenp != sizeof(lids_locks_t)) {
			lids_security_alert("Attempt to feed locks sysctl with garbage");
			return -EINVAL;
		}
		if (copy_from_user(&locks,buffer,sizeof(lids_locks_t)))
			return -EFAULT;
		if ((locks.magic1 != LIDS_MAGIC_1) || (locks.magic2 != LIDS_MAGIC_2) ||
		    (locks.magic3 != LIDS_MAGIC_3) || (locks.magic4 != LIDS_MAGIC_4)) {
			memset((char *)locks.passwd,'\0',sizeof(passwd_t));
			lids_security_alert("Attempt to feed locks sysctl bad magic numbers");
			return -EINVAL;
		}
		locks.passwd[sizeof(passwd_t)-1]=0; /* We don't take the risk */
		rmd160sig[0]=0;
		
#ifdef CONFIG_LIDS_ALLOW_SWITCH
		if ((!lids_first_time) || (locks.passwd[0])) {
			RMD((byte *)locks.passwd,hashcode);
			memset((char *)locks.passwd,'\0',sizeof(passwd_t));
			for (i=0; i<RMDsize/8; i++)
				sprintf(rmd160sig+2*i,"%02x", hashcode[i]);
		}
		if ( ((lids_first_time) && (!locks.passwd[0])) ||
		     (!strncmp(rmd160sig,lids_pw,LIDS_PW_LEN)) ) {
#else
		if ((lids_first_time) && (!locks.passwd[0])) {
#endif
			/* access granted ! */
			number_failed=0;
#ifdef CONFIG_LIDS_HANGUP
			if (current->tty!=NULL)
				set_bit(TTY_LIDS_NO_HANGUP, &(current->tty->flags));
#endif
			if (lids_process_flags(locks.flags)) {
				if (lids_first_time || lids_flag_raised(locks.flags,LIDS_FLAGS_RELOAD_CONF)) 
					lids_cap_val = lids_conf_cap;
				else
					lids_cap_val = locks.cap_bset;
				lids_log(0,"Changed: lids_cap_val=0x%x lids_flags=0x%x",lids_cap_val,lids_flags);
			}
			lids_first_time=0;
#ifdef CONFIG_LIDS_HANGUP
			if (current->tty!=NULL)
				clear_bit(TTY_LIDS_NO_HANGUP, &(current->tty->flags));
#endif
#ifdef CONFIG_LIDS_ALLOW_SWITCH
		}
		else {
			number_failed++;
			lids_security_alert("Give incorrect password (try #%d) with caps=0x%x and flags=0x%x",
					    number_failed,cap_t(locks.cap_bset),locks.flags);
			if (number_failed >= CONFIG_LIDS_MAX_TRY) {
				wait_after_fail=1;
				init_timer(&fail_timer);
				fail_timer.function=reenable_sysctl;
				fail_timer.data=(unsigned long)NULL;
				fail_timer.expires=jiffies+CONFIG_LIDS_TTW_FAIL*HZ;
				add_timer(&fail_timer);
			}
#else
			lids_security_alert("Attempt %d to switch caps/flags with caps=0x%x and flags=0x%x (feature disabled)",
					    number_failed,cap_t(locks.cap_bset),locks.flags);
#endif
		}
	}
	else {
		locks.cap_bset=lids_cap_val;
		locks.flags=lids_flags;
		LIDS_DBG("Sending caps=%#0x flags=%#0x\n",locks.cap_bset,locks.flags);
		if (*lenp < sizeof(lids_locks_t)) 
			return -EINVAL;
		if (copy_to_user(buffer,&locks,sizeof(lids_locks_t)))
			return -EFAULT;
	}
       return 0;
}

/*
 *	lids capability violate logging
 */ 

static char *lids_caps_desc[] =
{
	"CAP_CHOWN",    
	"CAP_DAC_OVERRIDE",     
	"CAP_DAC_READ_SEARCH", 
	"CAP_FOWNER",         
	"CAP_FSETID",        
	"CAP_KILL",    
	"CAP_SETGID",
	"CAP_SETUID",
	"CAP_SETPCAP", 
	"CAP_LINUX_IMMUTABLE", 
	"CAP_NET_BIND_SERVICE",
	"CAP_NET_BROADCAST",  
	"CAP_NET_ADMIN",     
	"CAP_NET_RAW",      
	"CAP_IPC_LOCK",    
	"CAP_IPC_OWNER",  
	"CAP_SYS_MODULE",
	"CAP_SYS_RAWIO", 
	"CAP_SYS_CHROOT", 
	"CAP_SYS_PTRACE", 
	"CAP_SYS_PACCT", 
	"CAP_SYS_ADMIN", 
	"CAP_SYS_BOOT",  
	"CAP_SYS_NICE",
	"CAP_SYS_RESOURCE",  
	"CAP_SYS_TIME",  
	"CAP_SYS_TTY_CONFIG",  
	"CAP_MKNOD",
	"CAP_LEASE",
	"CAP_HIDDEN",
	"CAP_KILL_PROTECTED",
	"CAP_PROTECTED",
	NULL
};

void  lids_cap_log(int cap)
{
	LIDS_DBG(":lids_cap =%lx, cap_effective=%x,cap=%d,lids_load=%d,lids_local_load=%d\n",current->lids_cap, current->cap_effective,cap,lids_load,lids_local_load);
	if (!cap_raised(lids_cap_val,cap))	
		lids_security_alert(" violated %s",lids_caps_desc[cap]);
}
